/********************************************************************************/
/*    H   H Y   Y TTTTT EEEEE  CCC       HYTEC ELECTRONICS LTD                  */
/*    H   H  Y Y    T   E     C          5 Cradock Road,                        */
/*    HHHHH   Y     T   EEE   C          Reading, Berks.    Tel: 0118 9757770	*/
/*    H   H   Y     T   E     C          RG2 0JT            Fax: 0118 9757566	*/
/*    H   H   Y     T   EEEEE  CCC       Web: www.hytec-electronics.co.uk       */
/********************************************************************************/
/********************************************************************************/
/*  __________________________________________________________________________  */
/*  | H Y T E C   8 0 0 1  32/64 B i t   D i g i t a l  A s y n  D r i v e r |  */
/*  --------------------------------------------------------------------------  */
/*                                                                              */
/*  Source file name      :-      drvHy8001Asyn.cpp                             */
/*                                                                              */
/*  Initial creation date :-      15-Aug-2012                                   */
/*                                                                              */
/*  Original Developers   :-      Jim Chen						                */
/*                                                                              */
/********************************************************************************/
/*                                                                              */
/*  Description :- This code is for Hytec 8001 digital EPICS support. It is     */
/*  based on asyn framework. It is osi complaint.		                    	*/
/*                                                                              */
/*  (C)2012 Hytec Electronics Ltd.                                              */
/*                                                                              */
/********************************************************************************/
/*                                                                              */
/* Modification record :-                                                       */
/*                                                                              */
/*  Date modified |  Person   | Version | Description of modification           */
/* ---------------|-----------|---------|-------------------------------------- */
/*    15/08/12    |  JC		  | 1.0     | Initial creation with new asyn model  */    
/*                |           |         |    		                            */    
/********************************************************************************/

#define DRIVER_VERSION	100

/******************/
/* Hytec Module includes */
/******************/
#include "drvHy8001Asyn.h"


/********************************************************************/
/*														           */
/* Public Sector												  */
/*													             */
/****************************************************************/
/* Constructor:
 *
 * Where: -
 *     portName = asyn port name
 *   carrierNum = carrier index number when adding a carrier card by ipacAddHy8002 call
 *    vectorNum = interrupt vector 
 *         dir  = 0: all inputs, 1: AB inputs/CD outputs, 2: AB outputs/CD inputs, 3: all outputs
 *   usememory  = 0: do not use memory for caching or update, 1: use memory for caching (input) or update (output)
 */ 
Hy8001Asyn::Hy8001Asyn(char *portName, int carrierNum, int vectorNum, int dir, int usememory)    
		: asynPortDriver(portName, MAX_CHANNELS, NUM_PARAMS, 
        asynInt32Mask | asynInt32ArrayMask | asynUInt32DigitalMask | asynDrvUserMask,
        asynInt32Mask | asynInt32ArrayMask | asynUInt32DigitalMask, 
        ASYN_MULTIDEVICE, 1, /* ASYN_CANBLOCK=0, ASYN_MULTIDEVICE=1, autoConnect=1 */
        0, 0)  /* Default priority and stack size */

{
    const char *hytecstr = "(Hytec Electronics Ltd., Reading, UK)";
    USHORT hytecID, modelID, issueNo;
    int ishytec;
    int ismodel;



	/* normal data channels definition */
    createParam(hy8001DataString, 			asynParamInt32, 	&hy8001Data_);

	/* controls definition */
    createParam(hy8001DirectionString, 		asynParamInt32, 	&hy8001Direction_);
    createParam(hy8001UseMemoryString, 		asynParamInt32, 	&hy8001UseMemory_);
    createParam(hy8001ClockSourceString, 	asynParamInt32, 	&hy8001ClockSource_);
    createParam(hy8001ScanRateString, 		asynParamInt32, 	&hy8001ScanRate_);
    createParam(hy8001DebRateString, 		asynParamInt32, 	&hy8001DebRate_);
    createParam(hy8001InhibitSourceString, 	asynParamInt32, 	&hy8001InhibitSource_);
    createParam(hy8001ContinuousString, 	asynParamInt32, 	&hy8001Continuous_);
    createParam(hy8001INTMaskLowString, 	asynParamInt32, 	&hy8001INTMaskLow_);
    createParam(hy8001INTMaskHighString, 	asynParamInt32, 	&hy8001INTMaskHigh_);
    createParam(hy8001DebMaskLowString, 	asynParamInt32, 	&hy8001DebMaskLow_);
    createParam(hy8001DebMaskHighString, 	asynParamInt32, 	&hy8001DebMaskHigh_);

    createParam(hy8001MemoryPointerString, 	asynParamInt32, 	&hy8001MemPointer_);
    createParam(hy8001FWVersionString, 		asynParamInt32, 	&hy8001FWVersion_);
    createParam(hy8001DriverVersionString,	asynParamInt32, 	&hy8001DriverVersion_);
    createParam(hy8001SupportString, 		asynParamInt32, 	&hy8001Support_);
	
	/* Save the parameters in the driver structure    */
    this->carrierNum		 = carrierNum;
    this->vectorNum          = vectorNum;
	
	setIntegerParam(hy8001Direction_, dir);
	setIntegerParam(hy8001UseMemory_, usememory);

	/* Map Register address spaces  */
	this->dio = (IO_MEMORY *) ipmBaseAddr(carrierNum, CARRIERSLOT, ipac_addrIO);
    if (this->dio == NULL)
    {  
        /* can't register card register address space */
    	printf("configHy8001: Unable to register IO space\n");
        return;
    }

    /* map the memory address spaces  */
    this->convMemoryBase = (USHORT *) ipmBaseAddr(carrierNum, CARRIERSLOT, ipac_addrMem);
    if (this->convMemoryBase == NULL)
    {  
        /* can't register card Memory address space */
     	printf("configHy8001: Unable to register memory space\n");
        return;
    }
    
	/* Compare the HY8001 prom header, manufacturer's ID and card model
     * number with the expected values. */
	hytecID = (this->dio->prom.hytecIdHigh << 8) | (this->dio->prom.hytecIdLow);
    ishytec = hytecID == HYTECID;
	modelID = (this->dio->prom.modelhigh << 8) | (this->dio->prom.modellow);
    ismodel = modelID == PROM_MODEL;
	issueNo = (int)this->dio->prom.pcbissue & 0xFF;

	/* check if we have been asked to be more chatty */
    /* start with some basic information */
    printf("PROM: Contents at offset - %#08x\n", (int)&this->dio->prom);
    printf("PROM: Manufacturer's ID  - 0x%04x %s\n", hytecID, ishytec ? hytecstr : "(Unknown)");
    printf("\n");
    /* output a bit more about this specific model and card */
    printf("PROM: Model #/revision   - %#04hx rev. %#04hx\n", modelID, issueNo);
    printf("\n");

	setIntegerParam(hy8001FWVersion_, issueNo);
	setIntegerParam(hy8001DriverVersion_, DRIVER_VERSION);

	if(this->dio->reg.csrcb & CSRCB_MEMMSIZE)
		this->memorysize = MIN_MEMSIZE / 2;								/* 64k samples */
	else
		this->memorysize = MAX_MEMSIZE / 2;								/* 256k samples */

	if(this->dio->reg.csrcb & CSRCB_HYTECIP)
		this->iscarrier = 1;											/* only IO */
	else
		this->iscarrier = 0;											/* is a carrier */

    /* check if everything passed okay */
    if (!ishytec)
	{
        printf("PROM: UNSUPPORTED MANUFACTURER ID; EXPECTED 0x%04x, %s\n", HYTECID, hytecstr);
//		return;															/* not hytec module, quit */
	}
    if (!ismodel)
	{
        printf("PROM: UNSUPPORTED MODEL NUMBER; EXPECTED %#04hx\n", PROM_MODEL);
//		return;															/* not hytec module, quit */
	}

    /* clear asyn message sent counters */
    this->messagesSent = 0;
    this->messagesFailed = 0;

    /* set up interrupt, work queue and mutex stuff */
    this->dio->reg.vector = vectorNum;            /* writes vector register offset 0x08 */
	if (ipmIntConnect(carrierNum, CARRIERSLOT, vectorNum, INTSRC, (int)this) < 0) {
      printf("configHy8001: failure connecting to interrupt\n");
      return;
    }

    this->intMsgQId = epicsMessageQueueCreate(MAX_MESSAGES, 
                                              (MAX_CHANNELS)*sizeof(int));
    if (epicsThreadCreate("Hy8001intQueuedTask",
                           epicsThreadPriorityLow,
                           epicsThreadGetStackSize(epicsThreadStackMedium),
                           (EPICSTHREADFUNC)intQueuedTaskC,
                           (void *)this) == NULL)
	{	
      	printf("configHy8001: failure creating message queue.\n");
      	return;
	}

    rebooting = 0;
 	epicsAtExit(rebootCallbackC, this);

}

/** Configuration extension command, called directly or from iocsh */
/* 
 * Where: -
 *     portName = asyn port name
 *  clocksource = 0: internal, 1: external from front panel, 2: external from strobe IN1
 *     scanrate = 0: 1kHz/divided by 1000, 1: 10kHz/devided by 100
 * 				  2: 100kHz/divided by 10, 3: 1MHz/divided by 1 
 *		debrate = 0: 100Hz/divided by 10000, 1: 200Hz/devided by 5000
 * 				  2: 500Hz/divided by 2000, 3: 1kHz/divided by 1000 
 *inhibitsource = 0: front panel, 1: strobe IN2, 2: strobe IN3, 3: strobe IN4
 *   continuous = 1: continuous updating, 0: single shot. For all output mode only
 *   intmasklow = interrupt mask for channel 0 ~ 31. Writing "1" to disable interrupt
 *  intmaskhigh = interrupt mask for channel 32 ~ 63. Writing "1" to disable interrupt
 *   debmasklow = debounce mask for channel 0 ~ 31. Writing "1" to enable debounce
 *  debmaskhigh = debounce mask for channel 32 ~ 63. Writing "1" to enable debounce
 */
int Hy8001Asyn::Hy8001ExtInit(int clocksource, int scanrate, int debrate, int inhibitsource, 
    		int continuous, unsigned int intmasklow, unsigned int intmaskhigh, 
			unsigned int debmasklow, unsigned int debmaskhigh)
{
	if((clocksource < 0) || (clocksource > 2))
	{
    	printf("configHy8001 error: Clock source setting %d wrong. Must be 0~2.\n", clocksource);
        return asynError;
	}
	if((scanrate < 0) || (scanrate > 3))
	{
    	printf("configHy8001 error: Scanning rate setting %d wrong. Must be 0~3.\n", scanrate);
        return asynError;
	}
	if((debrate < 0) || (debrate > 3))
	{
    	printf("configHy8001 error: Debounce rate setting %d wrong. Must be 0~3.\n", debrate);
        return asynError;
	}
	if((inhibitsource < 0) || (inhibitsource > 3))
	{
    	printf("configHy8001 error: Inhibit source setting %d wrong. Must be 0~3.\n", inhibitsource);
        return asynError;
	}

	setIntegerParam(hy8001ClockSource_, clocksource);
	setIntegerParam(hy8001ScanRate_, scanrate);
	setIntegerParam(hy8001DebRate_, debrate);
	setIntegerParam(hy8001InhibitSource_, inhibitsource);
	setIntegerParam(hy8001Continuous_, continuous);
	setIntegerParam(hy8001INTMaskLow_, intmasklow);
	setIntegerParam(hy8001INTMaskHigh_, intmaskhigh);
	setIntegerParam(hy8001DebMaskLow_, debmasklow);
	setIntegerParam(hy8001DebMaskHigh_, debmaskhigh);

	/*******************************************************************/
	/************* Working out real stuffs for ADC *********************/
	/*******************************************************************/

	setupIO();

	// tell listeners the changes
	callParamCallbacks();

	return(asynSuccess);
}


/*********************************************************/
/*                                                      */
/* Asyn interface Function implementations             */
/*                                                    */
/*****************************************************/

/* asynCommon methods */
void Hy8001Asyn::report(FILE *fp, int details)
{
	int dir, clocksource, scanrate, debrate, inhibitsource, continuous, intmasklow, intmaskhigh, debmasklow, debmaskhigh, fver, dver;

    fprintf(fp, "Port: %s, carrier slot %d, base address=%p\n", 
            this->portName, this->carrierNum, (void *)this->dio);
    if (details >= 1) 
    {
		getIntegerParam(hy8001Direction_, &dir);
		getIntegerParam(hy8001ClockSource_, &clocksource);
		getIntegerParam(hy8001ScanRate_, &scanrate);
		getIntegerParam(hy8001DebRate_, &debrate);
		getIntegerParam(hy8001InhibitSource_, &inhibitsource);
		getIntegerParam(hy8001Continuous_, &continuous);
		getIntegerParam(hy8001INTMaskLow_, &intmasklow);
		getIntegerParam(hy8001INTMaskHigh_, &intmaskhigh);
		getIntegerParam(hy8001DebMaskLow_, &debmasklow);
		getIntegerParam(hy8001DebMaskHigh_, &debmaskhigh);
		getIntegerParam(hy8001FWVersion_, &fver);
		getIntegerParam(hy8001DriverVersion_, &dver);

		
        fprintf(fp, "  csr CB=%x\n", this->dio->reg.csrcb);
        fprintf(fp, "  csr IO=%x\n", this->dio->reg.csrio);
        fprintf(fp, "  INT Vector=%x\n", this->vectorNum);
        fprintf(fp, "  dir(%d)=%s\n", dir, dir==0?"all inputs":(dir==1?"AB in/CD out":(dir==2?"AB out/CD in":"all outputs")));
        fprintf(fp, "  Clock source(%d)=%s\n", clocksource, clocksource==0?"Internal":(clocksource==1?"External front panel":"External Strobe IN1"));
        fprintf(fp, "  Scanning rate(%d)=%s\n", scanrate, scanrate==0?"1kHz/1000":(scanrate==1?"10kHz/100":(scanrate==2?"500Hz/2000":"1kHz/1000")));
        fprintf(fp, "  Debounce rate(%d)=%s\n", debrate, debrate==0?"100Hz/10000":(debrate==1?"200Hz/5000":(debrate==2?"500Hz/2000":"1kHz/1000")));
        fprintf(fp, "  Inhibit source(%d)=%s\n", inhibitsource, inhibitsource==0?"front panel":(inhibitsource==1?"strobe IN2":(inhibitsource==2?"strobe IN3":"strobe IN4")));
        fprintf(fp, "  Continuous update(%d)=%s\n", continuous, continuous==1?"yes":"no");
        fprintf(fp, "  Interrupt mask low=%x\n", intmasklow);
        fprintf(fp, "  Interrupt mask high=%x\n", intmaskhigh);
        fprintf(fp, "  Debounce mask low=%x\n", debmasklow);
        fprintf(fp, "  Debounce mask high=%x\n", debmaskhigh);
        fprintf(fp, "  Firmware Version=%d\n", fver);
        fprintf(fp, "  Driver Version=%d\n", dver);
        fprintf(fp, "  messages sent OK=%d; send failed (queue full)=%d\n",
                this->messagesSent, this->messagesFailed);       
    }
    asynPortDriver::report(fp, details);
}


/*****         asynInt32 methods       ******/
asynStatus Hy8001Asyn::writeInt32(asynUser *pasynUser, epicsInt32 value)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
	int function = pasynUser->reason;
	int dir;
	ULONG lastknown, data;

    this->getAddress(pasynUser, &channel);
	
    status = setIntegerParam(function, value);
	getIntegerParam(hy8001Direction_, &dir);

	/* decide what to do depending on which command we received */
	if (function == hy8001Data_)
	{
		if(this->iscarrier == 0)												/* if A and B sites are used for other IP cards */
		{
			if(channel > 31)
				printf("writeInt32 error: Try to write to non-digital IP card.\n");
			else if(dir == 0 || dir == 2)
				printf("writeInt32 error: Try to write to an input channel. Dir=%d\n", dir);				
			else if(channel < 16)
			{
				lastknown = this->dio->reg.lasthigh;
				data = 1 << channel;
				if(value == 0)
				{
					data = ~data;
					this->dio->reg.lasthigh = lastknown & data;
				}else
		        	this->dio->reg.lasthigh = lastknown | data;
			}
			else if(channel < 32)
			{
				lastknown = this->dio->reg.lasthighhigh;
				data = 1 << (channel - 16);
				if(value == 0)
				{
					data = ~data;
					this->dio->reg.lasthighhigh = lastknown & data;
				}else
		        	this->dio->reg.lasthighhigh = lastknown | data;
			}
		}else																	/* 64 digital IOs */
		{
			if((dir == 0) || ((dir == 1) && (channel<32)) || ((dir == 2) && (channel>=32)))
				printf("writeInt32 error: Try to write to an input channel(%d).\n", channel);
			else
			{
				if(channel < 16)
				{
					lastknown = this->dio->reg.lastlowlow;
					data = 1 << channel;
					if(value == 0)
					{
						data = ~data;
						this->dio->reg.lastlowlow = lastknown & data;
					}else
			        	this->dio->reg.lastlowlow = lastknown | data;
				}
				else if(channel < 32)
				{
					lastknown = this->dio->reg.lastlow;
					data = 1 << (channel - 16);
					if(value == 0)
					{
						data = ~data;
						this->dio->reg.lastlow = lastknown & data;
					}else
			        	this->dio->reg.lastlow = lastknown | data;
				}
				else if(channel < 48)
				{
					lastknown = this->dio->reg.lasthigh;
					data = 1 << (channel - 32);
					if(value == 0)
					{
						data = ~data;
						this->dio->reg.lasthigh = lastknown & data;
					}else
			        	this->dio->reg.lasthigh = lastknown | data;
				}
				else 
				{
					lastknown = this->dio->reg.lasthighhigh;
					data = 1 << (channel - 48);
					if(value == 0)
					{
						data = ~data;
						this->dio->reg.lasthighhigh = lastknown & data;
					}else
			        	this->dio->reg.lasthighhigh = lastknown | data;
				}
/*printf("ch=%d data=%x lknown=%x csrcb=%x csrio=%x rll=%x rl=%x rh=%x rhh=%x iscar=%d\n", 
	channel, data, lastknown, this->dio->reg.csrcb, this->dio->reg.csrio, 
	this->dio->reg.lastlowlow, this->dio->reg.lastlow, this->dio->reg.lasthigh, this->dio->reg.lasthighhigh, this->iscarrier);*/
			}
		}
	}
	else if (function == hy8001Direction_)
		changeDirection( value );
	else if (function == hy8001UseMemory_)
		useMemory(value);
	else if (function == hy8001ClockSource_)
        setClockSource( value );
	else if (function == hy8001ScanRate_)
        setScanRate( value );
	else if (function == hy8001DebRate_)
        setDenounceRate( value );
	else if (function == hy8001InhibitSource_)
        setInhibitSource( value );
	else if (function == hy8001Continuous_)
        setContinuous( value );
	else if (function == hy8001INTMaskLow_)
        setInterruptMask( value, 0 );
	else if (function == hy8001INTMaskHigh_)
        setInterruptMask( value, 1 );
	else if (function == hy8001DebMaskLow_)
        setDebounceMask( value, 0 );
	else if (function == hy8001DebMaskHigh_)
        setDebounceMask( value, 1 );

	// tell listeners the changes
  	callParamCallbacks();

	return( status );
}

asynStatus Hy8001Asyn::readInt32(asynUser *pasynUser, epicsInt32 *value)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
	int function = pasynUser->reason;
	int dir, pointer, ver;
	ULONG lastknown=0, data;

    if (this->rebooting) epicsThreadSuspendSelf();
   	
	getIntegerParam(hy8001Direction_, &dir);

	this->getAddress(pasynUser, &channel);
	
	if (function == hy8001Data_)
	{
		if(this->iscarrier == 0)												/* if A and B sites are used for other IP cards */
		{
			if(channel > 31)
				printf("writeInt32 error: Try to read from non-digital IP card.\n");
			else if(channel < 16)
				lastknown = this->dio->reg.lasthigh;
			else if(channel < 32)
				lastknown = this->dio->reg.lasthighhigh;

			data = 1 << (channel - 16);
			*value = (lastknown & data) != 0 ? 1 : 0;
		}else																	/* 64 digital IOs */
		{
			if(channel < 16)
				lastknown = this->dio->reg.lastlowlow;
			else if(channel < 32)
				lastknown = this->dio->reg.lastlow;
			else if(channel < 48)
				lastknown = this->dio->reg.lasthigh;
			else 
				lastknown = this->dio->reg.lasthighhigh;
				
			data = 1 << channel;
			*value = (lastknown & data) != 0 ? 1 : 0;
		}
	}

	else if(function == hy8001Support_)
        *value = DRIVER_SUPPORT;                                                /* driver support */
	else if(function == hy8001MemPointer_)
	{
		pointer = (this->dio->reg.mempointerlow + this->dio->reg.mempointerhigh * 65536)/2;
		setIntegerParam(hy8001MemPointer_, pointer);
        *value = pointer;                                                		/* overflow */
	}
	else if(function == hy8001FWVersion_)
	{
		getIntegerParam(hy8001FWVersion_, &ver);
        *value = ver;                                               	 		/* firmware version */		
	}
	else if(function == hy8001DriverVersion_)
	{
		getIntegerParam(hy8001DriverVersion_, &ver);
        *value = ver;                                               	 		/* driver version */		
    }else
	{
		asynPortDriver::readInt32(pasynUser, value);
	}
	
  	callParamCallbacks();
	return( status );
}

asynStatus Hy8001Asyn::getBounds(asynUser *pasynUser, epicsInt32 *low, epicsInt32 *high)
{
	*low = 0;
	*high = 63;
	/* not usually implemented... See asyn documentation for details if needed. */
	return( asynSuccess );
}

/* asynInt32Array methods */
asynStatus Hy8001Asyn::readInt32Array(asynUser *pasynUser, epicsInt32 *data, size_t nelements, size_t *nactual)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
	int function = pasynUser->reason;
	int dir, use, pointer;
	unsigned int i;
	USHORT mask;

    if (this->rebooting) epicsThreadSuspendSelf();

	getIntegerParam(hy8001Direction_, &dir);
	getIntegerParam(hy8001UseMemory_, &use);

	this->getAddress(pasynUser, &channel);
	
    if(function == hy8001Data_)
	{
		if((dir != 3) && (use))
		{
			pointer = (this->dio->reg.mempointerlow + this->dio->reg.mempointerhigh * 65536) * 2;	// convert to word
			for(i=0; i<nelements; i++)
			{
				if(dir == 0)
				{
					if(channel < 16){														// Slot A
						mask = 1 << channel;
						*(data+i) = (epicsInt32)((((USHORT)this->convMemoryBase[pointer+3]) & mask) != 0 ? 1 : 0);
					}else if(channel < 32){													// Slot B
						mask = 1 << (channel - 16);
						*(data+i) = (epicsInt32)((((USHORT)this->convMemoryBase[pointer+2]) & mask) != 0 ? 1 : 0);
					}else if(channel < 48){													// Slot C
						mask = 1 << (channel - 32);
						*(data+i) = (epicsInt32)((((USHORT)this->convMemoryBase[pointer+1]) & mask) != 0 ? 1 : 0);
					}else{																	// Slot D
						mask = 1 << (channel - 48);
						*(data+i) = (epicsInt32)((((USHORT)this->convMemoryBase[pointer]) & mask) != 0 ? 1 : 0);
					}
				}
				if(dir == 1)
				{
					if(channel < 16){														// Slot A
						mask = 1 << channel;
						*(data+i) = (epicsInt32)((((USHORT)this->convMemoryBase[pointer+3]) & mask) != 0 ? 1 : 0);
					}else if(channel < 32){													// Slot B
						mask = 1 << (channel - 16);
						*(data+i) = (epicsInt32)((((USHORT)this->convMemoryBase[pointer+2]) & mask) != 0 ? 1 : 0);
					}
				}
				if(dir == 2)
				{
					if(channel < 16){														// Slot C
						mask = 1 << channel;
						*(data+i) = (epicsInt32)((((USHORT)this->convMemoryBase[pointer+1]) & mask) != 0 ? 1 : 0);
					}else if(channel < 32){													// Slot D
						mask = 1 << (channel - 16);
						*(data+i) = (epicsInt32)((((USHORT)this->convMemoryBase[pointer]) & mask) != 0 ? 1 : 0);
					}
				}
				pointer = pointer - 4;
				if(pointer < 0) pointer = this->memorysize * 4 - 4;
			}
			*nactual = nelements;
		}
    }


	status = asynSuccess;
	return( status );
}

asynStatus Hy8001Asyn::writeInt32Array(asynUser *pasynUser, epicsInt32 *value, size_t nElements)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
	int function = pasynUser->reason;
	int dir, use;
	size_t i;
	USHORT low, high;

    if (this->rebooting) epicsThreadSuspendSelf();

	getIntegerParam(hy8001Direction_, &dir);
	getIntegerParam(hy8001UseMemory_, &use);

	this->getAddress(pasynUser, &channel);
	
    if(function == hy8001Data_)
	{
		if((dir != 0) && (use))
		{
			for(i=0; i<nElements; i++)
			{
				if(dir == 3)
				{
					low = (USHORT)((*(value+i)) & 0xFFFF);
					high = (USHORT)((*(value+i)) / 65536);
					if(i/2 == 0)
					{
						this->convMemoryBase[i*4+3] = low;									// slot A
						this->convMemoryBase[i*4+2] = high;									// slot B
					}else
					{
						this->convMemoryBase[i*4+1] = low;									// slot C
						this->convMemoryBase[i*4] = high;									// slot D
					}
				}
				if(dir == 1)
				{
					low = (USHORT)((*(value+i)) & 0xFFFF);
					high = (USHORT)((*(value+i)) / 65536);
					this->convMemoryBase[i*4+1] = low;										// slot C
					this->convMemoryBase[i*4] = high;										// slot D
				}
				if(dir == 2)
				{
					low = (USHORT)((*(value+i)) & 0xFFFF);
					high = (USHORT)((*(value+i)) / 65536);
					this->convMemoryBase[i*4+3] = low;										// slot A
					this->convMemoryBase[i*4+2] = high;										// slot B
				}
			}
		}
    }

	status = asynSuccess;
	return( status );
}



asynStatus Hy8001Asyn::readUInt32Digital(asynUser *pasynUser, epicsUInt32 *value, epicsUInt32 mask)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
	int function = pasynUser->reason;
	int dir;

    if (this->rebooting) epicsThreadSuspendSelf();

	getIntegerParam(hy8001Direction_, &dir);

	this->getAddress(pasynUser, &channel);

    if(function == hy8001Data_)
	{
		if(this->iscarrier == 0)															/* if A and B sites are used for other IP cards */
		{
			if(channel > 1)
				printf("writeInt32 error: Try to read from non-digital IP card.\n");
			else if(channel == 0)
				*value = (epicsUInt32)(this->dio->reg.lasthigh & mask);
			else
				*value = (epicsUInt32)(this->dio->reg.lasthighhigh & mask);
		}else																				/* 64 digital IOs */
		{
			if(channel == 0)
				*value = (epicsUInt32)(this->dio->reg.lastlowlow & mask);
			else if(channel == 1) 
				*value = (epicsUInt32)(this->dio->reg.lastlow & mask);
			else if(channel == 2) 
				*value = (epicsUInt32)(this->dio->reg.lasthigh & mask);
			else 
				*value = (epicsUInt32)(this->dio->reg.lasthighhigh & mask);
		}		
    }

	status = asynSuccess;
	return( status );
}

asynStatus Hy8001Asyn::writeUInt32Digital(asynUser *pasynUser, epicsUInt32 value, epicsUInt32 mask)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
	int function = pasynUser->reason;
	int dir;

    if (this->rebooting) epicsThreadSuspendSelf();

	getIntegerParam(hy8001Direction_, &dir);

	this->getAddress(pasynUser, &channel);
	
    if(function == hy8001Data_)
	{
		if(this->iscarrier == 0)															/* if A and B sites are used for other IP cards */
		{
			if(channel > 0)
				printf("writeInt32 error: Try to read from non-digital IP card.\n");
			else if(channel == 0)
				this->dio->reg.lasthigh = value & mask;										// write to C
			else
				this->dio->reg.lasthighhigh = value & mask;									// write to D
		}else																				/* 64 digital IOs */
		{
			if(channel == 0)
				this->dio->reg.lastlowlow = value & mask;									// write to A
			else if(channel == 1)
				this->dio->reg.lastlow = value & mask;										// write to B
			else if(channel == 2)
				this->dio->reg.lasthigh = value & mask;										// write to C
			else
				this->dio->reg.lasthighhigh = value & mask;									// write to D
		}
    }

	status = asynSuccess;
	return( status );
}

/*********************************************************/
/*                                                      */
/* Private Sector                                      */
/*                                                    */
/*****************************************************/

void Hy8001Asyn::setupIO()
{
	int dir, use, source, rate, debrate, inhsource, mode, masklow, maskhigh, debmasklow, debmaskhigh;
	
	if (this->rebooting)  epicsThreadSuspendSelf();

	getIntegerParam(hy8001Direction_, &dir);
	getIntegerParam(hy8001UseMemory_, &use);
	getIntegerParam(hy8001ClockSource_, &source);
	getIntegerParam(hy8001ScanRate_, &rate);
	getIntegerParam(hy8001DebRate_, &debrate);
	getIntegerParam(hy8001InhibitSource_, &inhsource);
	getIntegerParam(hy8001Continuous_, &mode);
	getIntegerParam(hy8001INTMaskLow_, &masklow);
	getIntegerParam(hy8001INTMaskHigh_, &maskhigh);
	getIntegerParam(hy8001DebMaskLow_, &debmasklow);
	getIntegerParam(hy8001DebMaskHigh_, &debmaskhigh);
/* this first write to the CSR sets this bit and clears all others */
    this->dio->reg.csrio = 0;				                             		/* clear interrupt */   

/* clear pointer to zero at start */
    this->dio->reg.mempointerlow = 0;
    this->dio->reg.mempointerhigh = 0;

/* initialise direction */
   	changeDirection( dir );	                             						/* check average setting overflow */

/* initialise whether use memory for caching or updating */
	useMemory(use);

/* set up clock source */
	setClockSource( source );	          

/* set up scanning rate */
	setScanRate( rate );	              

/* set up debounce rate */
    setDenounceRate( debrate );

/* set up inhibit source */
	setInhibitSource( inhsource );
	
/* set up continuous mode or single shot for output */
	setContinuous( mode );

/* set up interrupt and debounce masks */
    setInterruptMask( masklow, 0 );
    setInterruptMask( maskhigh, 1 );
    setDebounceMask( debmasklow, 0 );
    setDebounceMask( debmaskhigh, 1 );
	if(dir != 3) this->dio->reg.csrio|= CSRIO_SCANEN;							/* if there are inputs, go! */
    ipmIrqCmd(this->carrierNum, CARRIERSLOT, 0, ipac_irqEnable);				/* enable interrupt */

/* get the last status */
	lastKnown[0] = this->dio->reg.lastlowlow;
	lastKnown[1] = this->dio->reg.lastlow;
	lastKnown[2] = this->dio->reg.lasthigh;
	lastKnown[3] = this->dio->reg.lasthighhigh;
}

int Hy8001Asyn::changeDirection( int dir  )
{
	USHORT csrio = this->dio->reg.csrio;
    if (this->rebooting) epicsThreadSuspendSelf();

	csrio &= ~(3<<CSRIO_DIRSHIFT);												// clear direction first 
	csrio |= dir<<CSRIO_DIRSHIFT;												// set direction
	
	this->dio->reg.csrio = csrio;												// write to csr
    return( asynSuccess );

}

int Hy8001Asyn::useMemory( int use  )
{
	USHORT csrio = this->dio->reg.csrio;
    if (this->rebooting) epicsThreadSuspendSelf();

	if(use)
		csrio |= CSRIO_MEMEN;													// use memory for caching or update 
	else
		csrio &= ~CSRIO_MEMEN;													// disable memory for caching or update
	
	this->dio->reg.csrio = csrio;												// write to csr
    return( asynSuccess );

}
int Hy8001Asyn::setClockSource( int source )
{
	USHORT csrio = this->dio->reg.csrio;
    if (this->rebooting) epicsThreadSuspendSelf();

	csrio &= ~(3<<CSRIO_CLKSHIFT);												// clear clock source setting first 
	csrio |= source<<CSRIO_CLKSHIFT;											// set clock source
	
	this->dio->reg.csrio = csrio;												// write to csr
    return( asynSuccess );

}

int Hy8001Asyn::setScanRate( int rate )
{
    if (this->rebooting) epicsThreadSuspendSelf();
	USHORT csrio = this->dio->reg.csrio;

	csrio &= ~(3<<CSRIO_SCANSHIFT);												// clear scan rate setting first 
	csrio |= rate<<CSRIO_SCANSHIFT;												// set scan rate
	
	this->dio->reg.csrio = csrio;												// write to csr
    return( asynSuccess );

}

int Hy8001Asyn::setDenounceRate(int debrate )
{
	USHORT csrio = this->dio->reg.csrio;
    if (this->rebooting) epicsThreadSuspendSelf();

	csrio &= ~(3<<CSRIO_DEBSHIFT);												// clear debounce rate setting first 
	csrio |= debrate<<CSRIO_DEBSHIFT;											// set debounce rate
	
	this->dio->reg.csrio = csrio;												// write to csr
    return( asynSuccess );

}


int Hy8001Asyn::setInhibitSource( int inhsource )
{
	USHORT csrio = this->dio->reg.csrio;
    if (this->rebooting) epicsThreadSuspendSelf();

	csrio &= ~(3<<CSRIO_INHSHIFT);												// clear inhibit setting first 
	csrio |= inhsource<<CSRIO_INHSHIFT;											// set inhibit
	
	this->dio->reg.csrio = csrio;												// write to csr
    return( asynSuccess );

}

int Hy8001Asyn::setContinuous( int mode )
{
	int dir;
	USHORT csrio = this->dio->reg.csrio;
    if (this->rebooting) epicsThreadSuspendSelf();

	getIntegerParam(hy8001Direction_, &dir);
	if(dir == 0) return asynError;												// cannot do when it is all inputs

	if(mode)
		csrio |= CSRIO_RINGMODE;												// set ring mode
	else
		csrio &= ~CSRIO_RINGMODE;												// clear ring mode, single shot 
	
	this->dio->reg.csrio = csrio;												// write to csr
    return( asynSuccess );
}

int Hy8001Asyn::setInterruptMask( unsigned int mask, int highlow )
{
	USHORT masklow, maskhigh;
    if (this->rebooting) epicsThreadSuspendSelf();

	masklow = mask & 0xFFFF;
	maskhigh = (mask & 0xFFFF0000) >> 16;

	if(highlow)	
	{	
		this->dio->reg.maskhigh = masklow;
		this->dio->reg.maskhighhigh = maskhigh;
	}else
	{
		this->dio->reg.masklowlow = masklow;
		this->dio->reg.masklow = maskhigh;
	}
    return( asynSuccess );

}

int Hy8001Asyn::setDebounceMask( unsigned int mask, int highlow )
{
	USHORT masklow, maskhigh;
    if (this->rebooting) epicsThreadSuspendSelf();

	masklow = mask & 0xFFFF;
	maskhigh = (mask & 0xFFFF0000) >> 16;

	if(highlow)	
	{	
		this->dio->reg.debhigh = masklow;
		this->dio->reg.debhighhigh = maskhigh;
	}else
	{
		this->dio->reg.deblowlow = masklow;
		this->dio->reg.deblow = maskhigh;
	}
    return( asynSuccess );
}

void Hy8001Asyn::rebootCallback()
{
    this->dio->reg.csrio &= ~CSRIO_SCANEN;                          				/* Clear the scan bit  */
    this->rebooting = 1;				                            				/* flag we are rebooting */
}

int Hy8001Asyn::getMemoryData(int card, USHORT mask, int startpoint)
{
	int i;
	for(i=0;i<this->memorysize;i++)
	{
		if(card == 0)																// Slot A
			this->rawData[i] = (epicsInt32)((((USHORT)this->convMemoryBase[startpoint+3]) & mask) != 0 ? 1 : 0);
		else if(card == 1)															// Slot B
			this->rawData[i] = (epicsInt32)((((USHORT)this->convMemoryBase[startpoint+2]) & mask) != 0 ? 1 : 0);
		else if(card == 2)															// Slot C
			this->rawData[i] = (epicsInt32)((((USHORT)this->convMemoryBase[startpoint+1]) & mask) != 0 ? 1 : 0);
		else																		// Slot D
			this->rawData[i] = (epicsInt32)((((USHORT)this->convMemoryBase[startpoint]) & mask) != 0 ? 1 : 0);

		startpoint = startpoint - 4;
		if(startpoint <= 0) startpoint = this->memorysize*4 - 4;
	}
	return( asynSuccess );	
}


/*********************************************************/
/*                                                      */
/* External Wrapped Routines      					   */
/*                                                    */
/*****************************************************/

// These functions must have C linkage because they are called from other EPICS components
extern "C" {
static void rebootCallbackC(void *drvPvt)
{
    Hy8001Asyn *pPvt = (Hy8001Asyn *)drvPvt;
    pPvt->rebootCallback();
}

static void INTSRC(int drvPvt)
{
    Hy8001Asyn *pPvt = (Hy8001Asyn *)drvPvt;
    pPvt->intSR();
}

static void intQueuedTaskC(void *drvPvt)
{
    Hy8001Asyn *pPvt = (Hy8001Asyn *)drvPvt;
    pPvt->intQueuedTask();	
}

}

/*********************************************************/
/*                                                      */
/* Interrupt and Queue Task and external Routines      */
/*                                                    */
/*****************************************************/
void Hy8001Asyn::intSR()
{
    unsigned int data[6];

	useMemory(0);
	ipmIrqCmd(carrierNum, CARRIERSLOT, 0, ipac_irqDisable);								/* disable interrupt */

    data[0] = this->dio->reg.mempointerlow;                                             /* memory pointer low */
    data[1] = this->dio->reg.mempointerhigh;                                            /* memory pointer high */
    data[2] = this->dio->reg.lastlowlow;                                            	/* channel 0~15 */
    data[3] = this->dio->reg.lastlow;                                            		/* channel 16~31 */
    data[4] = this->dio->reg.lasthigh;                                            		/* channel 32~47 */
    data[5] = this->dio->reg.lasthighhigh;                                            	/* channel 48~63 */
    
	if(this->dio->reg.csrio & CSRIO_COSBIT) 
	{
		this->dio->reg.csrio &= ~CSRIO_COSBIT;											/* clear interrupt */
    	/* Wake up queue task which calls callback routines if it is the correct interrupt.*/
	    if (epicsMessageQueueTrySend(this->intMsgQId, data, sizeof(data)) == 0)
    	   	messagesSent++;
    	else
    	   	messagesFailed++;
	}
}

void Hy8001Asyn::intQueuedTask()
{
    unsigned int data[6];
	int dir,chan,card, start;
	USHORT digdata=0, mask = 1, value;

    while(1) {
	    /* Wait for event from interrupt routine */
	    epicsMessageQueueReceive(intMsgQId, data, sizeof(data));
                 
	  	getIntegerParam(hy8001Direction_, &dir);

	    start = (data[0] + data[1]*65536)*2;                                             	/* interrupted memory pointer in WORD */

		if(dir == 0) 
		{
			for(card=0;card<4;card++)
			{
				digdata = data[card+2] ^ lastKnown[card];
				mask = 1;
		       	for (chan=0; chan<16; chan++)
    		  	{
					if((digdata & mask) != 0)
					{
						value = (data[card+2] & mask) != 0 ? 1 : 0;
    					setIntegerParam(chan+card*16, hy8001Data_, value);
						callParamCallbacks(chan+card*16);
						// get waveform data
						getMemoryData(card, mask, start);
						doCallbacksInt32Array((epicsInt32 *) &rawData,
                        		this->memorysize, hy8001Data_, chan+card*16);						
					}
					mask = mask << 1;
				}
				lastKnown[card] = data[card+2];
			}
        }else if(dir == 1)
		{
			for(card=0;card<2;card++)
			{
				digdata = data[card+2] ^ lastKnown[card];
				mask = 1;
		       	for (chan=0; chan<16; chan++)
    		  	{
					if((digdata & mask) != 0)
					{
						value = (data[card+2] & mask) != 0 ? 1 : 0;
    					setIntegerParam(chan+card*16, hy8001Data_, value);
						callParamCallbacks(chan+card*16);
						// get waveform data
						getMemoryData(card, mask, start);
						doCallbacksInt32Array((epicsInt32 *) &rawData,
                        		this->memorysize, hy8001Data_, chan+card*16);						
					}
					mask = mask << 1;
				}
				lastKnown[card] = data[card+2];
			}			
		}else if(dir == 2)
		{
			for(card=0;card<2;card++)
			{
				digdata = data[card+4] ^ lastKnown[card+2];
				mask = 1;
		       	for (chan=0; chan<16; chan++)
    		  	{
					if((digdata & mask) != 0)
					{
						value = (data[card+4] & mask) != 0 ? 1 : 0;
    					setIntegerParam(chan+card*16, hy8001Data_, value);
						callParamCallbacks(chan+card*16);
						// get waveform data
						getMemoryData(card+2, mask, start);
						doCallbacksInt32Array((epicsInt32 *) &rawData,
                        		this->memorysize, hy8001Data_, chan+card*16);						
					}
					mask = mask << 1;
				}
				lastKnown[card+2] = data[card+4];
			}			
		}

	    ipmIrqCmd(carrierNum, CARRIERSLOT, 0, ipac_irqEnable);
	    callParamCallbacks();
	}
}

