/*******************/
/* System includes */
/*******************/

#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <limits.h>

/******************/
/* EPICS includes */
/******************/

/* EPICS includes */
#include <drvIpac.h>
#include <iocsh.h>
#include <epicsExport.h>
#include <epicsThread.h>
#include <epicsString.h>
#include <epicsTimer.h>
#include <epicsExit.h>
#include <epicsMutex.h>
#include <epicsMessageQueue.h>
#include <cantProceed.h>
#include <devLib.h>
#include <errlog.h>
#include <asynPortDriver.h>

#ifndef DRVHY8001IPASYN_H
#define DRVHY8001IPASYN_H
#endif

/**************************************************************/
/* A couple of key parameters defining the memory layout and channel numbers. */
#define MAX_MEMSIZE  			524288         	/* maximum memory size: 512K x 32bit*/
#define MIN_MEMSIZE  			131072         	/* minimum memory size: 128K x 32bit*/
#define MAX_MESSAGES  			100           	/* maximum number of messages */
#define MAX_CHANNELS    		64      		/* maximum number of digital channels on 8001 card */
#define MIN_CHANNELS    		32      		/* minimum number of digital channels on 8001 card */

/* Identifiers in PROM. */
#define HYTECID		  			0x8003
#define PROM_MODEL  			0x8001

/* define individual bits in the carrier board's CSR CB register*/
#define CSRCB_RESET   	0x0001
#define CSRCB_INTEN   	0x0002
#define CSRCB_MEMMODE 	0x0040
#define CSRCB_FRONTINH 	0x0080
#define CSRCB_IPAINT 	0x0100
#define CSRCB_IPBINT   	0x0200
#define CSRCB_INTREL   	0x0400
#define CSRCB_MEMMSIZE 	0x0800
#define CSRCB_HYTECIP 	0x1000

#define CSRCB_T012_8301	0x0000
#define CSRCB_T012_8302	0x0000
#define CSRCB_T012_8303	0xA000
#define CSRCB_T012_8304	0xE000
#define CSRCB_T012_8305	0xC000

#define CSRCB_INTSELSHIFT 	2

/* define individual bits in the carrier board's CSR IO register*/
#define CSRIO_SCANEN   			0x0001
#define CSRIO_MEMEN  			0x0200
#define CSRIO_RINGMODE 			0x1000
#define CSRIO_INHSTATE 			0x2000
#define CSRIO_INTEN 			0x4000
#define CSRIO_COSBIT 			0x8000

#define CSRIO_CLKSHIFT  		1
#define CSRIO_DEBSHIFT  		3
#define CSRIO_INHSHIFT  		7
#define CSRIO_SCANSHIFT 		5
#define CSRIO_MEMINHSHIFT		1
#define CSRIO_DIRSHIFT			10

#define CARRIERSLOT 			2               /* carrier IO equivalent slot */

#define INIT 					0
#define RUNTIME 				1

/* driver support flags */
/* Bit0 - continuos mode support */
/* Bit1 - trigger mode support */
#define DRIVER_SUPPORT     		0x0003   		/* support continuous and trigger modes */

#define WAITING_TRIGGER         0           	/* waiting for trigger */
#define TRIGGERRED_COLLECTING   1           	/* triggerred and collecting data */
#define FINISHED_AVAILABLE      2           	/* finished and data available */

#ifndef USHORT 
#define USHORT unsigned short
#endif
#ifndef ULONG
#define ULONG unsigned int
#endif

/*****************************************************************************/
/*                                                                           */
/*                              Data Structures                              */
/*                                                                           */
/*****************************************************************************/


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* 8001 Register and PROM Layout                                              */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* The I/O memory is divided into two areas: register and PROM. We separate
 * them out for coding convenience. 
 */

/* This data structure maps the I/O registers */
typedef struct
{
    USHORT vector;                  /* 0x00 */
    USHORT unused;                 	/* 0x02 */
    USHORT memoff;          		/* 0x04 */
    USHORT unused1;               	/* 0x06 */
    USHORT csrcb;              		/* 0x08 */
    USHORT unused2[3];              /* 0x0A~0x0E */
    USHORT ipstatus;               	/* 0x10 */
    USHORT unused3;               	/* 0x12 */
    USHORT csrio;               	/* 0x14 */
    USHORT unused4;               	/* 0x16 */
    USHORT masklowlow;             	/* 0x18 */
    USHORT masklow;               	/* 0x1A */
    USHORT maskhigh;               	/* 0x1C */
    USHORT maskhighhigh;           	/* 0x1E */
    USHORT deblowlow;             	/* 0x20 */
    USHORT deblow;               	/* 0x22 */
    USHORT debhigh;               	/* 0x24 */
    USHORT debhighhigh;           	/* 0x26 */
    USHORT lastlow;             	/* 0x28 site B */
    USHORT lastlowlow;           	/* 0x2A site A */
    USHORT lasthighhigh;           	/* 0x2C site D */
    USHORT lasthigh;       		   	/* 0x2E site C */
    USHORT mempointerhigh;          /* 0x30 */
    USHORT mempointerlow;         	/* 0x32 */
    USHORT sequencecount;          	/* 0x34 */
} REGISTERS;


/* This data structure maps the image of the ID PROM.  The PROM
 * starts at offset 0x80 from the base of the register memory. */
typedef struct
{
    USHORT notinterested1[21];      /* 0x00~0x29 */
    USHORT hytecIdHigh;             /* 0x2A */
	USHORT notinterested2;			/* 0x2C */
    USHORT hytecIdLow;              /* 0x2E */
	USHORT notinterested3;			/* 0x30 */
    USHORT modelhigh;               /* 0x32 */
	USHORT notinterested4;			/* 0x34 */
    USHORT modellow;                /* 0x36 */
    USHORT notinterested5[5];       /* 0x38~0x40 */
    USHORT pcbissue;                /* 0x42 */
	USHORT notinterested6;			/* 0x44 */
    USHORT Xilinxpcbissue;          /* 0x46 */
    USHORT notinterested7;          /* 0x48 */
} ID_PROM;

/* Now put them together, the I/O memory map. */
typedef struct
{
    REGISTERS reg;                              /* 0x00 - 0x31 */
    char unused[0x200-sizeof(REGISTERS)];       /* 0x32 - 0x1FF */
    ID_PROM prom;                               /* 0x200 - 0x249 */
} IO_MEMORY;

/* enumerated type of the driver's operational modes */
typedef enum{		
	hy8001Carrier,
	hy8001Digital
} hy8001Modes;


/*****************************************************************************/
/*                                                                           */
/*                      Interrupt/Queue Routines	                         */
/*                                                                           */
/*****************************************************************************/

extern "C" {
static void rebootCallbackC     ( void *drvPvt );
static void INTSRC              ( int drvPvt ); 	/* Interrupt Service Routine */
static void intQueuedTaskC      ( void *drvPvt );
}

/** This is the class definition for the Hy8001Asyn class
  */
class Hy8001Asyn : public asynPortDriver {
public:
	Hy8001Asyn(char *portName, int carrierNum, int vectorNum, int dir, int usememory);

    /* These are the methods that we override from asynPortDriver */
    virtual asynStatus writeInt32(asynUser *pasynUser, epicsInt32 value);
    virtual asynStatus readInt32(asynUser *pasynUser, epicsInt32 *value);
    virtual asynStatus readUInt32Digital(asynUser *pasynUser, epicsUInt32 *value, epicsUInt32 mask);
    virtual asynStatus writeUInt32Digital(asynUser *pasynUser, epicsUInt32 value, epicsUInt32 mask);
    virtual asynStatus readInt32Array(asynUser *pasynUser, epicsInt32 *value, size_t nElements, size_t *nIn);
    virtual asynStatus writeInt32Array(asynUser *pasynUser, epicsInt32 *value, size_t nElements);
    virtual asynStatus getBounds(asynUser *pasynUser, epicsInt32 *low, epicsInt32 *high);
    virtual void report(FILE *fp, int details);
    
    /* these are specific to Hy8001Asyn class */
    int Hy8001ExtInit(int clocksource, int scanrate, int debrate, int inhibitsource, 
    		int continuous, unsigned int intmasklow, unsigned int intmaskhigh, 
			unsigned int debmasklow, unsigned int debmaskhigh);	
    void rebootCallback();
	void intSR(); 				/* Interrupt Service Routine */
	void intQueuedTask();

protected:
    int hy8001Data_;
    #define FIRST_8001_PARAM hy8001Data_
	int hy8001Direction_;
	int hy8001UseMemory_;
	int hy8001ClockSource_;
	int hy8001ScanRate_;
	int hy8001DebRate_;
	int hy8001InhibitSource_;
	int hy8001Continuous_;
	int hy8001INTMaskLow_;
	int hy8001INTMaskHigh_;
	int hy8001DebMaskLow_;
	int hy8001DebMaskHigh_;

	int hy8001MemPointer_;
	int hy8001FWVersion_;
	int hy8001DriverVersion_;
	int hy8001Support_;
    #define LAST_8001_PARAM hy8001Support_

private:
	void setupIO               	();
	int changeDirection			( int dir );
	int useMemory				( int use );
	int setClockSource          ( int source );
	int setScanRate             ( int rate );
	int setDenounceRate         ( int rate );
	int setInhibitSource		( int source );
	int setContinuous	 		( int mode );
	int setInterruptMask		( unsigned int mask, int highlow );
	int setDebounceMask			( unsigned int mask, int highlow );

	int writebit				( int bit, int value );
	int write32bit				( unsigned int mask, unsigned int value );
	int readbit					( int bit, int *value );
	int read32bit				( unsigned int mask, unsigned int value );
	int getMemoryData			( int card, USHORT mask, int startpoint );
    
	volatile IO_MEMORY 	*dio;			   			/* pointer to registers space, ID */ 
	volatile USHORT 	*convMemoryBase;   			/* pointer to memory space */

    ULONG 	rawData[MAX_MEMSIZE/2];					/* memory data mapping */

	/* Storage for the current Configuration parameters */

    USHORT	carrierNum;		                		/* carrier index number when adding carrier */
    USHORT	vectorNum;		                		/* Interrupt vector number - boot time configuration only */

	int		memorysize;
	int		iscarrier;								/* determine the board is a carrier or just IO */
	USHORT 	lastKnown[4];							/* last known status of all 64 bit */
	
/* driver specific parameters */
    int		rebooting;		                		/* Flag to indicate is the IOC is currently re-booting */

/* Work Queue Stuff */
    epicsMessageQueueId intMsgQId;
    int messagesSent;                       		/* for report */
    int messagesFailed;                     		/* for report */
    epicsMessageQueueId intSweepQId;
};

#define hy8001DataString	        "DATA"
#define hy8001DirectionString	    "DIRECTION"
#define hy8001UseMemoryString	    "USEMEMORY"
#define hy8001ClockSourceString	    "CLOCK_SOURCE"
#define hy8001ScanRateString	    "SCAN_RATE"
#define hy8001DebRateString	    	"DEBRATE"
#define hy8001InhibitSourceString	"INHIBIT_SOURCE"
#define hy8001ContinuousString		"CONTINUOUS"
#define hy8001INTMaskLowString		"INTMASKLOW"
#define hy8001INTMaskHighString		"INTMASKHIGH"
#define hy8001DebMaskLowString		"DEBMASKLOW"
#define hy8001DebMaskHighString		"DEBMASKHIGH"

#define hy8001MemoryPointerString	"MEMPOINTER"
#define hy8001FWVersionString		"FWVERSION"
#define hy8001DriverVersionString	"DRIVERVERSION"
#define hy8001SupportString			"SUPPORT"

#define NUM_PARAMS (&LAST_8001_PARAM - &FIRST_8001_PARAM + 1)

/** Configuration command, called directly or from iocsh */
/* 
 * Where: -
 *     portName = asyn port name
 *   carrierNum = carrier index number when adding a carrier card by ipacAddHy8001 call
 *    vectorNum = interrupt vector 
 *         dir  = 0: all inputs, 1: AB inputs/CD outputs, 2: AB outputs/CD inputs, 3: all outputs
 *   usememory  = 0: do not use memory for caching or update, 1: use memory for caching (input) or update (output)
 */
extern "C" int Hy8001AsynInit(char *portName, int vmeSlotNum, int vectorNum, 
					int dir, int usememory)
{
    Hy8001Asyn *pHy8001Asyn = new Hy8001Asyn(portName, vmeSlotNum, vectorNum, dir, usememory);
    pHy8001Asyn = NULL;  /* This is just to avoid compiler warnings */
    return(asynSuccess);
}

/** Configuration extension command, called directly or from iocsh */
/* 
 * Where: -
 *     portName = asyn port name
 *  clocksource = 0: internal, 1: external from front panel, 2: external from strobe IN1
 *     scanrate = 0: 1kHz/divided by 1000, 1: 10kHz/devided by 100
 * 				  2: 100kHz/divided by 10, 3: 1MHz/divided by 1 
 *		debrate = 0: 100Hz/divided by 10000, 1: 200Hz/devided by 5000
 * 				  2: 500Hz/divided by 2000, 3: 1kHz/divided by 1000 
 *inhibitsource = 0: front panel, 1: strobe IN2, 2: strobe IN3, 3: strobe IN4
 *   continuous = 1: continuous updating, 0: single shot. For all output mode only
 *   intmasklow = interrupt mask for channel 0 ~ 31. Writing "1" to disable interrupt
 *  intmaskhigh = interrupt mask for channel 32 ~ 63. Writing "1" to disable interrupt
 *   debmasklow = debounce mask for channel 0 ~ 31. Writing "1" to enable debounce
 *  debmaskhigh = debounce mask for channel 32 ~ 63. Writing "1" to enable debounce
 */
extern "C" int Hy8001AsynExtInit(char *portName, int clocksource, int scanrate, int debrate, int inhibitsource, 
    		int continuous, unsigned int intmasklow, unsigned int intmaskhigh, 
			unsigned int debmasklow, unsigned int debmaskhigh)
{
    asynStatus status;
    asynInterface *pasynInterface;
    asynUser *pasynUser;
    Hy8001Asyn *pHy8001 = NULL;

    pasynUser = pasynManager->createAsynUser(0, 0);
    status = pasynManager->connectDevice(pasynUser, portName, 0);
    if (status != asynSuccess) {
        errlogPrintf("Hy8001AsynExtInit, error in connectDevice %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pasynInterface = pasynManager->findInterface(pasynUser, asynInt32Type, 1);
    if (!pasynInterface) {
        errlogPrintf("Hy8001AsynExtInit cannot find Int32 interface %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pHy8001 = (Hy8001Asyn*) pasynInterface->drvPvt;
    pHy8001->Hy8001ExtInit(clocksource, scanrate, debrate, inhibitsource, continuous, intmasklow, intmaskhigh,
			debmasklow, debmaskhigh);
    return(asynSuccess);
} 
				
// Init
static const iocshArg initArg0 = { "Port name",iocshArgString};
static const iocshArg initArg1 = { "vmeSlotNum",iocshArgInt};
static const iocshArg initArg2 = { "vectorNum",iocshArgInt};
static const iocshArg initArg3 = { "Direction", iocshArgInt};
static const iocshArg initArg4 = { "Usememory", iocshArgInt};
static const iocshArg * const initArgs[] = {&initArg0,
                                             &initArg1,
                                             &initArg2,
                                             &initArg3,
                                             &initArg4};
static const iocshFuncDef initFuncDef = {"Hy8001AsynInit",5,initArgs};
static void initCallFunc(const iocshArgBuf *args)
{
    Hy8001AsynInit(args[0].sval, args[1].ival, args[2].ival, args[3].ival, args[4].ival);
}

// Extension init
static const iocshArg initArg0Ext = { "Port name",iocshArgString};
static const iocshArg initArg1Ext = { "clock source",iocshArgInt};
static const iocshArg initArg2Ext = { "scan rate",iocshArgInt};
static const iocshArg initArg3Ext = { "debounce rate",iocshArgInt};
static const iocshArg initArg4Ext = { "inhibit source",iocshArgInt};
static const iocshArg initArg5Ext = { "continuous update", iocshArgInt};
static const iocshArg initArg6Ext = { "interrupt mask low", iocshArgInt};
static const iocshArg initArg7Ext = { "interrupt mask high", iocshArgInt};
static const iocshArg initArg8Ext = { "debounce mask low",iocshArgInt};
static const iocshArg initArg9Ext = { "debounce mask high",iocshArgInt};
static const iocshArg * const initArgsExt[] = {&initArg0Ext,
                                             &initArg1Ext,
                                             &initArg2Ext,
                                             &initArg3Ext,
                                             &initArg4Ext,
                                             &initArg5Ext,
                                             &initArg6Ext,
                                             &initArg7Ext,
                                             &initArg8Ext,
                                             &initArg9Ext};
static const iocshFuncDef initFuncDefExt = {"Hy8001AsynExtInit",10,initArgsExt};
static void initCallFuncExt(const iocshArgBuf *args)
{
    Hy8001AsynExtInit(args[0].sval, args[1].ival, args[2].ival, args[3].ival,
					args[4].ival, args[5].ival, args[6].ival, args[7].ival,
					args[8].ival, args[9].ival);
}

void drvHy8001AsynRegister(void)
{
    iocshRegister(&initFuncDef,initCallFunc);
    iocshRegister(&initFuncDefExt,initCallFuncExt);
}

extern "C" {
epicsExportRegistrar(drvHy8001AsynRegister);
}

